import datetime

import pytest

from contessa.db import Connector
from contessa.models import TIME_FILTER_DEFAULT
from contessa.utils import AggregatedResult
from test.conftest import FakedDatetime

from contessa.models import (
    create_default_check_class,
    ResultTable,
    DQBase,
    QualityCheck,
)


@pytest.fixture(scope="module")
def results():
    return AggregatedResult(total_records=5, passed=3, failed=2,)


def test_quality_check_init_row(rule, results, conn: Connector):
    DQBase.metadata.clear()
    qc = create_default_check_class(
        ResultTable(
            schema_name="data_quality", table_name="booking", model_cls=QualityCheck
        )
    )
    assert qc.__tablename__ == "quality_check_booking"
    assert qc.__name__ == "DataQualityQualityCheckBooking"
    t = datetime.datetime(2019, 8, 10, 10, 0, 0)

    qc.__table__.create(conn.engine)
    instance = qc()
    instance.init_row(rule, results, conn, context={"task_ts": t})

    assert instance.task_ts == t
    assert instance.attribute == "src"
    assert instance.rule_name == "not_null_name"
    assert instance.rule_type == "not_null"
    assert instance.rule_description == "True when data is null."
    assert instance.total_records == 5
    assert instance.failed == 2
    assert instance.passed == 3
    assert instance.failed_percentage == 40
    assert instance.passed_percentage == 60
    assert instance.median_30_day_failed is None
    assert instance.median_30_day_passed is None
    assert instance.time_filter == TIME_FILTER_DEFAULT
    assert instance.status == "invalid"


def test_set_medians(conn: Connector, monkeypatch):
    DQBase.metadata.clear()
    qc = create_default_check_class(
        ResultTable(schema_name="data_quality", table_name="t", model_cls=QualityCheck)
    )
    qc.__table__.create(conn.engine)
    instance = qc()

    conn.execute(
        """
        insert into data_quality.quality_check_t(attribute, rule_name, rule_type, failed, passed, task_ts, time_filter)
        values
          ('a', 'b', 'not_null', 10, 200, '2018-09-11T13:00:00', 'not_set'),
          ('a', 'b', 'not_null', 3, 22, '2018-09-10T13:00:00', 'not_set'),
          ('a', 'b', 'not_null', 11, 110, '2018-09-09T13:00:00', 'not_set'),
          ('a', 'b', 'not_null', 55, 476, '2018-09-08T13:00:00', 'not_set'),
          ('a', 'b', 'not_null', 77, 309, '2018-07-12T13:00:00', 'not_set') -- should not be taken
    """
    )

    monkeypatch.setattr("contessa.models.datetime", FakedDatetime)
    instance.set_medians(conn)

    assert instance.median_30_day_failed == 10.5
    assert instance.median_30_day_passed == 155



import operator

import pytest

from contessa.normalizer import RuleNormalizer


@pytest.mark.parametrize(
    "rules_def, normalized",
    [
        (
            [
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "columns": ["a", "b", "c"],
                    "separate_time_filters": ["c", "u"],
                }
            ],
            [
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "a",
                    "time_filter": "c",
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "b",
                    "time_filter": "c",
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "c",
                    "time_filter": "c",
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "a",
                    "time_filter": "u",
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "b",
                    "time_filter": "u",
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "c",
                    "time_filter": "u",
                },
            ],
        ),
        (
            [{"name": "not_null_name", "type": "not_null", "columns": ["a", "b"]}],
            [
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "a",
                    "time_filter": None,
                },
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "b",
                    "time_filter": None,
                },
            ],
        ),
        (
            [
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "a",
                    "time_filter": "a",
                }
            ],
            [
                {
                    "name": "not_null_name",
                    "type": "not_null",
                    "column": "a",
                    "time_filter": "a",
                }
            ],
        ),
    ],
)
def test_normalizer(rules_def, normalized):
    results = RuleNormalizer().normalize(rules_def)
    op = operator.itemgetter("name", "type", "column", "time_filter")
    expected = sorted(normalized, key=op)
    results_sorted = sorted(results, key=op)
    assert expected == results_sorted


def test_normalizer_separate_time_filters_bad_format():
    rules_def = [
        {"type": "not_null", "column": "a", "separate_time_filters": [{"column": "c"}]}
    ]
    with pytest.raises(ValueError):
        results = RuleNormalizer().normalize(rules_def)


@pytest.mark.parametrize(
    "rules_def, normalized",
    [
        (
            [
                {
                    "type": "not_null",
                    "column": "a",
                    "separate_time_filters": [
                        {"column": "a", "days": 10},
                        {"column": "b"},
                    ],
                }
            ],
            [
                {
                    "type": "not_null",
                    "column": "a",
                    "time_filter": [{"column": "a", "days": 10}],
                },
                {"type": "not_null", "column": "a", "time_filter": [{"column": "b"}]},
            ],
        ),
        (
            [
                {
                    "type": "not_null",
                    "column": "a",
                    "separate_time_filters": [
                        {"column": "a", "days": 10},
                        {"column": "b", "days": 5},
                    ],
                }
            ],
            [
                {
                    "type": "not_null",
                    "column": "a",
                    "time_filter": [{"column": "a", "days": 10}],
                },
                {
                    "type": "not_null",
                    "column": "a",
                    "time_filter": [{"column": "b", "days": 5}],
                },
            ],
        ),
    ],
)
def test_normalizer_separate_time_filters(rules_def, normalized):
    def sort_rules(rule):
        return (
            rule["type"],
            rule["column"],
            rule["time_filter"][0]["column"]
            if isinstance(rule["time_filter"], list)
            else rule["time_filter"]["column"],
        )

    results = RuleNormalizer().normalize(rules_def)
    expected = sorted(normalized, key=sort_rules)
    results_sorted = sorted(results, key=sort_rules)
    assert expected == results_sorted


@pytest.mark.parametrize(
    "rules_def, normalized",
    [
        (
            [
                {
                    "type": "not_null",
                    "columns": ["a", "b", "c"],
                    "condition": "d IN ('start', 'stay_end', 'stopover_end')",
                    "time_filter": None,
                }
            ],
            [
                {
                    "type": "not_null",
                    "column": "a",
                    "condition": "d IN ('start', 'stay_end', 'stopover_end')",
                    "time_filter": None,
                },
                {
                    "type": "not_null",
                    "column": "b",
                    "condition": "d IN ('start', 'stay_end', 'stopover_end')",
                    "time_filter": None,
                },
                {
                    "type": "not_null",
                    "column": "c",
                    "condition": "d IN ('start', 'stay_end', 'stopover_end')",
                    "time_filter": None,
                },
            ],
        )
    ],
)
def test_normalizer_condition(rules_def, normalized):
    results = RuleNormalizer().normalize(rules_def)
    op = operator.itemgetter("type", "column", "condition")
    expected = sorted(normalized, key=op)
    results_sorted = sorted(results, key=op)
    assert expected == results_sorted



from datetime import datetime, timedelta
from statistics import median
from typing import Dict, Any
import json

from sqlalchemy import and_, Column, DateTime, MetaData, text, UniqueConstraint
from sqlalchemy.dialects.postgresql import (
    BIGINT,
    DOUBLE_PRECISION,
    INTEGER,
    TEXT,
    TIMESTAMP,
)
from sqlalchemy.ext.declarative import (
    AbstractConcreteBase,
    declarative_base,
    declared_attr,
)

from contessa.base_rules import Rule
from contessa.db import Connector
from contessa.utils import AggregatedResult

# default schema for results is `data_quality`, but it can be overridden by passing
# ResultTable to runner. constructing concrete model for QualityCheck (that's abstract) will
# swap the schemas before creation - see `create_default_quality_check_class`
DQBase = declarative_base(metadata=MetaData(schema="data_quality"))

TIME_FILTER_DEFAULT = "not_set"


class QualityCheck(AbstractConcreteBase, DQBase):
    """
    Representation of abstract quality check table.
    """

    __abstract__ = True
    _table_prefix = "quality_check"

    id = Column(BIGINT, primary_key=True)
    attribute = Column(TEXT, nullable=False)
    rule_name = Column(TEXT, nullable=False)
    rule_type = Column(TEXT, nullable=False)
    rule_description = Column(TEXT)
    total_records = Column(INTEGER)

    failed = Column(INTEGER)
    median_30_day_failed = Column(DOUBLE_PRECISION)
    failed_percentage = Column(DOUBLE_PRECISION)

    passed = Column(INTEGER)
    median_30_day_passed = Column(DOUBLE_PRECISION)
    passed_percentage = Column(DOUBLE_PRECISION)

    status = Column(TEXT)
    time_filter = Column(
        TEXT,
        default=TIME_FILTER_DEFAULT,
        server_default=TIME_FILTER_DEFAULT,
        nullable=False,
    )
    task_ts = Column(TIMESTAMP(timezone=True), nullable=False, index=True)
    created_at = Column(
        DateTime(timezone=True),
        server_default=text("NOW()"),
        nullable=False,
        index=True,
    )

    @declared_attr
    def __table_args__(cls):
        """
        Concrete classes derived from this abstract one should have unique check among the columns
        that below. But the constraint needs to have unique name, therefore we are using
        @declared_attr here to construct name of the constraint using its table name.
        :return:
        """
        return (
            UniqueConstraint(
                "attribute",
                "rule_name",
                "rule_type",
                "task_ts",
                "time_filter",
                name=f"{cls.__tablename__}_unique",
            ),
        )

    def init_row(
        self,
        rule: Rule,
        results: AggregatedResult,
        conn: Connector,
        context: Dict = None,
    ):
        """
        Count metrics we want to measure and set them to quality check object.
        """

        # todo - add to doc
        self.task_ts = context["task_ts"]
        self.attribute = rule.attribute
        self.rule_name = rule.name
        self.rule_type = rule.type
        self.rule_description = rule.description

        self.total_records = results.total_records
        self.failed = results.failed
        self.passed = results.passed

        self.set_medians(conn)

        if rule.time_filter:
            self.time_filter = str(rule.time_filter)
        else:
            self.time_filter = TIME_FILTER_DEFAULT
        self.failed_percentage = self._perc(self.failed, self.total_records)
        self.passed_percentage = self._perc(self.passed, self.total_records)
        self.status = "invalid" if self.failed > 0 else "valid"

    def _perc(self, a, b):
        res = 0
        try:
            res = (a / b) * 100
        except ZeroDivisionError:
            pass
        return res

    def set_medians(self, conn: Connector, days=30):
        """
        Calculate median of passed/failed quality checks from last 30 days.
        """
        now = datetime.today().date()
        past = now - timedelta(days=days)
        cls = self.__class__

        session = conn.make_session()
        checks = (
            session.query(cls.failed, cls.passed)
            .filter(and_(cls.task_ts <= str(now), cls.task_ts >= str(past)))
            .all()
        )
        session.expunge_all()
        session.commit()
        session.close()

        failed = [ch.failed for ch in checks]
        self.median_30_day_failed = median(failed) if failed else None

        passed = [ch.passed for ch in checks]
        self.median_30_day_passed = median(passed) if passed else None

    def __repr__(self):
        return f"Rule ({self.attribute} - {self.rule_name} - {self.rule_type} - {self.task_ts})"


class ConsistencyCheck(AbstractConcreteBase, DQBase):
    """
    Representation of abstract consistency check table.
    """

    __abstract__ = True
    _table_prefix = "consistency_check"

    id = Column(BIGINT, primary_key=True)
    type = Column(TEXT, nullable=False)
    name = Column(TEXT, nullable=False)
    description = Column(TEXT)
    left_table = Column(TEXT, nullable=False)
    right_table = Column(TEXT, nullable=False)

    status = Column(TEXT)
    time_filter = Column(
        TEXT,
        nullable=False,
        default=TIME_FILTER_DEFAULT,
        server_default=TIME_FILTER_DEFAULT,
    )
    task_ts = Column(TIMESTAMP(timezone=True), nullable=False, index=True)
    created_at = Column(
        DateTime(timezone=True),
        server_default=text("NOW()"),
        nullable=False,
        index=True,
    )

    @declared_attr
    def __table_args__(cls):
        """
        Concrete classes derived from this abstract one should have unique check among the columns
        that below. But the constraint needs to have unique name, therefore we are using
        @declared_attr here to construct name of the constraint using its table name.
        :return:
        """
        return (
            UniqueConstraint(
                "type",
                "name",
                "left_table",
                "right_table",
                "task_ts",
                "time_filter",
                name=f"{cls.__tablename__}_unique",
            ),
        )

    def init_row(
        self,
        check: Dict,
        results: AggregatedResult,
        left_table_name: str,
        right_table_name: str,
        time_filter=None,
        context: Dict = None,
        **_,
    ):
        """
        Set result to consistency check object.
        """
        self.type = check["type"]
        self.task_ts = context["task_ts"]
        self.name = check["name"]
        self.description = check["description"]
        self.left_table = left_table_name
        self.right_table = right_table_name
        if isinstance(time_filter, str):
            self.time_filter = time_filter
        else:
            self.time_filter = json.dumps(time_filter)
        self.status = "valid" if results.failed == 0 else "invalid"

    def __repr__(self):
        return f"Rule ({self.type} - {self.name} - {self.task_ts})"


class Table:
    def __init__(self, schema_name, table_name):
        self.schema_name = schema_name
        self.table_name = table_name

    @property
    def fullname(self):
        # todo - solve sanitization. problem - can be already sanitized
        return f"{self.schema_name}.{self.table_name}"


class ResultTable(Table):
    """
    Name will be constructed as "quality_check_{load_table_name}", e.g. "quality_check_my_table".
    Can be overridden by `table_name`.
    """

    def __init__(self, schema_name, table_name, model_cls):
        self.model_cls = model_cls
        prefix = model_cls._table_prefix
        table_name = f"{prefix}_{table_name}"
        super().__init__(schema_name, table_name)

    def to_camel_case(self, snake_str):
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    @property
    def clsname(self):
        """
        Construct a name for dynamic creation of cls for quality table.
        Should be unique in runtime.
        """
        name = super().fullname.replace(".", "_")
        camel_case = self.to_camel_case(name)
        return camel_case[0].title() + camel_case[1:]


def create_default_check_class(result_table: ResultTable):
    """
    This will construct type/class (not object) that will have special name that its prefixed
    with its table. It's better because sqlalchemy can yell somethings if we would use same class
    for generic quality checks.
    Basically, its somethings as:

        class MyTableQualityCheck(QualityCheck):
            id = ...
            ...

    But it has dynamic name - MyTable is replaced for the table we are doing quality check for.
    :return: class with dynamically created name
    """

    attributedict = {
        "__tablename__": result_table.table_name,
        "id": Column(BIGINT, primary_key=True),
        "__mapper_args__": {
            "polymorphic_identity": result_table.table_name,
            "concrete": True,
        },
    }
    cls = type(result_table.clsname, (result_table.model_cls,), attributedict)
    cls.metadata.schema = result_table.schema_name
    cls.__table__.schema = result_table.schema_name
    return cls


class CheckResult:
    rule_name: str
    rule_type: str
    rule_description: str
    total_records: int
    failed: int
    passed: int
    time_filter: str
    failed_percentage: float
    passed_percentage: float
    status: str
    failed_example: Any
    context: Dict

    def init_row(
        self,
        rule: Rule,
        results: AggregatedResult,
        conn: Connector,
        context: Dict = None,
    ):
        self.rule_name = rule.name
        self.rule_type = rule.type
        self.rule_description = rule.description

        self.total_records = results.total_records
        self.failed = results.failed
        self.passed = results.passed
        self.failed_example = results.failed_example

        if rule.time_filter:
            self.time_filter = str(rule.time_filter)
        self.failed_percentage = self._perc(self.failed, self.total_records)
        self.passed_percentage = self._perc(self.passed, self.total_records)
        self.status = "invalid" if self.failed > 0 else "valid"
        self.context = context

    def init_row_consistency(
        self,
        check: Dict,
        results: AggregatedResult,
        left_table_name,
        right_table_name,
        time_filter=None,
        context: Dict = None,
    ):
        self.rule_type = check["type"]
        self.rule_name = check["name"]
        self.rule_description = check["description"]

        self.total_records = results.total_records
        self.failed = results.failed
        self.passed = results.passed
        self.failed_example = results.failed_example

        if time_filter:
            self.time_filter = time_filter
        self.failed_percentage = self._perc(abs(self.failed), self.total_records)
        self.passed_percentage = self._perc(self.passed, self.total_records)
        self.status = "valid" if results.failed == 0 else "invalid"

        context.update(
            {"left_table_name": left_table_name, "right_table_name": right_table_name}
        )
        self.context = context

    def _perc(self, a, b):
        res = 0
        try:
            res = (a / b) * 100
        except ZeroDivisionError:
            pass
        return res



"""set_defaults_and_nullables

Revision ID: a179e5ca0ad2
Revises: 480e6618700d
Create Date: 2020-01-13 13:43:44.679965

"""
from typing import List

from alembic import op
import sqlalchemy as sa
from sqlalchemy import create_engine, inspect, Table, MetaData

from sqlalchemy.dialects.postgresql import TEXT
from contessa.models import QualityCheck, ConsistencyCheck, TIME_FILTER_DEFAULT

# revision identifiers, used by Alembic.
revision = "a179e5ca0ad2"
down_revision = "480e6618700d"
branch_labels = None
depends_on = None

config = None


def get_config():
    global config
    if config:
        return config

    from alembic import context

    config = context.config

    return config


def get(name):
    return get_config().get_main_option(name)


def get_quality_tables(table_prefix) -> List[str]:
    url = get("sqlalchemy.url")
    schema = get("schema")

    engine = create_engine(url)
    inspector = inspect(engine)

    all_tables = inspector.get_table_names(schema=schema)
    quality_tables = [x for x in all_tables if x.startswith(table_prefix)]

    return quality_tables


def set_time_filter_to_none(schema_name, table_name):
    qc_table = Table(
        table_name,
        MetaData(schema=schema_name),
        sa.Column("time_filter", TEXT),
        # Other columns not needed for the data migration
    )
    op.execute(
        qc_table.update()
        .where(qc_table.c.time_filter == TIME_FILTER_DEFAULT)
        .values({"time_filter": None})
    )


def set_time_filter_to_default(schema_name, table_name):
    qc_table = Table(
        table_name,
        MetaData(schema=schema_name),
        sa.Column("time_filter", TEXT),
        # Other columns not needed for the data migration
    )
    op.execute(
        qc_table.update()
        .where(qc_table.c.time_filter.is_(None))
        .values({"time_filter": TIME_FILTER_DEFAULT})
    )


def upgrade():
    schema = get("schema")
    not_null = lambda t, col: op.alter_column(t, col, nullable=False, schema=schema)

    print("Migration Quality Check")
    for table_name in get_quality_tables(QualityCheck._table_prefix):
        print(f"Migrate table {table_name}")
        not_null(table_name, "attribute")
        not_null(table_name, "rule_name")
        not_null(table_name, "rule_type")

        set_time_filter_to_default(schema, table_name)
        op.alter_column(
            table_name,
            "time_filter",
            nullable=False,
            server_default=TIME_FILTER_DEFAULT,
            schema=schema,
        )

    print("Migration Consistency Check")
    for table_name in get_quality_tables(ConsistencyCheck._table_prefix):
        print(f"Migrate table {table_name}")
        not_null(table_name, "type")
        not_null(table_name, "name")
        not_null(table_name, "left_table")
        not_null(table_name, "right_table")

        set_time_filter_to_default(schema, table_name)
        op.alter_column(
            table_name,
            "time_filter",
            nullable=False,
            server_default=TIME_FILTER_DEFAULT,
            schema=schema,
        )


def downgrade():
    schema = get("schema")
    null = lambda t, col: op.alter_column(t, col, nullable=True, schema=schema)

    for table_name in get_quality_tables(QualityCheck._table_prefix):
        print(f"Migrate table {table_name}")
        null(table_name, "attribute")
        null(table_name, "rule_name")
        null(table_name, "rule_type")

        op.alter_column(
            table_name, "time_filter", nullable=True, server_default=None, schema=schema
        )
        set_time_filter_to_none(schema, table_name)

    for table_name in get_quality_tables(ConsistencyCheck._table_prefix):
        print(f"Migrate table {table_name}")
        null(table_name, "type")
        null(table_name, "name")
        null(table_name, "left_table")
        null(table_name, "right_table")

        op.alter_column(
            table_name, "time_filter", nullable=True, server_default=None, schema=schema
        )
        set_time_filter_to_none(schema, table_name)



from contessa.db import Connector
from contessa.models import DQBase
from test.conftest import FakedDatetime
from test.integration.conftest import TEST_DB_URI
import unittest
from unittest import mock

from contessa import ConsistencyChecker


class TestReturnResults(unittest.TestCase):
    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.left_table_name = "raw_booking"
        self.right_table_name = "booking"
        self.ts_nodash = (
            FakedDatetime.now().isoformat().replace("-", "").replace(":", "")
        )
        self.now = FakedDatetime.now()

        sql = [
            "DROP SCHEMA if exists tmp CASCADE;",
            "CREATE SCHEMA IF NOT EXISTS tmp;",
            "CREATE SCHEMA IF NOT EXISTS hello;",
            f"""
                    CREATE TABLE IF NOT EXISTS tmp.{self.left_table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                    CREATE TABLE IF NOT EXISTS hello.{self.right_table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                INSERT INTO tmp.{self.left_table_name}
                    (src, dst, price, turnover_after_refunds, initial_price, created_at)
                VALUES
                    ('BTS', NULL, 1, 100, 11, '2018-09-12T11:50:00'),
                    (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T15:50:00'),
                    ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T11:50:00'),
                    ('VIE', 'VIE', 4, 0.0, 0.0, '2018-09-11T11:50:00')
                """,
            f"""
                INSERT INTO hello.{self.right_table_name}
                    (src, dst, price, turnover_after_refunds, initial_price, created_at)
                VALUES
                    ('BTS', NULL, 1, 100, 11, '2018-09-12T11:50:00'),
                    (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T15:50:00'),
                    ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T11:50:00')
            """,
        ]
        self.conn = Connector(TEST_DB_URI)
        for s in sql:
            self.conn.execute(s)

        self.consistency_checker = ConsistencyChecker(TEST_DB_URI)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema tmp CASCADE;")
        self.conn.execute(f"DROP schema hello CASCADE;")
        DQBase.metadata.clear()

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_consistency(self):
        result = self.consistency_checker.run(
            self.consistency_checker.COUNT,
            left_check_table={"schema_name": "tmp", "table_name": self.left_table_name},
            right_check_table={
                "schema_name": "hello",
                "table_name": self.right_table_name,
            },
            context={"task_ts": self.now},
        )

        self.assertEqual(result.status, "invalid")
        self.assertEqual(result.context["left_table_name"], "tmp.raw_booking")
        self.assertEqual(result.context["right_table_name"], "hello.booking")



from contessa import ContessaRunner
from contessa.executor import refresh_executors
from contessa.models import Table
from test.utils import normalize_str
from contessa.rules import SqlRule


def test_rule_context_formatted_in_where():
    class TestRule(SqlRule):
        @property
        def sql(self):
            return "select a, b, c from {{table_fullname}}_{{ts_nodash}}"

    r = TestRule(
        name="test_rule_name",
        type="test_rule_type",
        condition="created_at >= '{{ts_nodash}}'::timestamptz - interval '10 minutes'",
        description="Greater than 0 when bags <> 0",
    )
    check_table = Table("raw", "table")
    context = ContessaRunner.get_context(check_table, {"ts_nodash": "20190101T000000"})

    # executor holds context of run, so set it
    refresh_executors(check_table, "", context)

    result = r.sql_with_where
    expected = """
		select a, b, c
		from raw.table_20190101T000000
		where created_at >= '20190101T000000'::timestamptz - interval '10 minutes'
	"""
    assert normalize_str(result) == normalize_str(expected)



from datetime import timedelta, datetime
from dataclasses import dataclass
from enum import Enum
from typing import Union, List, Optional, Dict


@dataclass
class TimeFilterColumn:
    column: str
    since: Optional[Union[timedelta, datetime]] = None
    since_inclusive: bool = True
    until: Optional[Union[timedelta, datetime, str]] = None
    until_inclusive: bool = False

    def compose_sql(self, now: datetime):
        result = "("
        if self.since:
            since_str = self.time_to_sql_value(self.since, now)
            result += (
                f"{self.column} >{'=' if self.since_inclusive else ''} {since_str}"
            )
        if self.since and self.until:
            result += " AND "
        if self.until:
            until_str = self.time_to_sql_value(self.until, now)
            result += (
                f"{self.column} <{'=' if self.until_inclusive else ''} {until_str}"
            )
        result += ")"
        return result

    def time_to_sql_value(self, time, now):
        if isinstance(time, str):
            if time == "now":
                time = now
            else:
                raise ValueError("'now' is only allowed string value")
        if isinstance(time, timedelta):
            time = now - time
        return f"'{time.strftime('%Y-%m-%d %H:%M:%S UTC')}'::timestamptz"

    def __str__(self):
        if self.since and self.until:
            return f"{self.column} between {self.since} and {self.until}"
        elif self.since:
            return f"{self.column} > {self.since}"
        elif self.until:
            return f"{self.column} < {self.until}"
        else:
            raise ValueError(
                "Incorrect configuration, at least one of 'since' or 'until' has to be set"
            )


class TimeFilterConjunction(Enum):
    AND = "AND"
    OR = "OR"


@dataclass()
class TimeFilter:
    columns: List[TimeFilterColumn]
    conjunction: TimeFilterConjunction = TimeFilterConjunction.OR
    now: datetime = None

    def __post_init__(self):
        self.now = datetime.now()

    def __str__(self):
        c = f" {self.conjunction.value.lower()} "
        s = c.join(str(c) for c in self.columns)
        if self.now:
            s += f" relative to {self.now}"
        return f"<TimeFilter {s}>"

    @property
    def sql(self):
        sep = f" {self.conjunction.value} "
        return sep.join(c.compose_sql(self.now) for c in self.columns)


# for backwards compatibility with previous versions
def parse_time_filter(time_filter: Union[str, List[Dict], TimeFilter]) -> TimeFilter:
    default_since = timedelta(days=30)

    if time_filter is None or isinstance(time_filter, TimeFilter):
        return time_filter
    if isinstance(time_filter, str):
        return TimeFilter(
            columns=[TimeFilterColumn(time_filter, since=default_since, until="now")]
        )
    if isinstance(time_filter, list):
        return TimeFilter(
            columns=[
                TimeFilterColumn(
                    column["column"],
                    since=timedelta(days=column.get("days", default_since.days)),
                    until="now",
                )
                for column in time_filter
            ]
        )



import abc
from typing import Union, List, Dict, Optional
from contessa.time_filter import TimeFilter, parse_time_filter


class Rule(metaclass=abc.ABCMeta):
    """
    Representation of one rule.
    Method apply define how it will be evaluated using the executor that will inject all the
    needed attributes to apply it (df, connector etc.).

    Attributes:
        executor_cls    Executor, indication of which executor class can execute this rule
        description     str, description of the rule

    :param name: str
    :param type: str
    :param description: str
    :param time_filter: str
    :param condition: str
    """

    executor_cls = None
    description = None

    def __init__(
        self,
        name,
        type,
        description,
        time_filter: Optional[Union[str, List[Dict], TimeFilter]] = None,
        condition=None,
    ):
        self.name = name
        self.type = type
        self.time_filter = parse_time_filter(time_filter)
        self.condition = condition
        self.description = description

    @property
    def attribute(self):
        """What attributes(columns) is this rule using."""
        return None

    def apply(self, **kwargs):
        """
        Override this method to check a specific rule.
        """
        raise NotImplementedError

    def __str__(self):
        tf = f" - {self.time_filter}" or ""
        return f"Rule {self.name} of type {self.type}{tf}"



import pytest
from sqlalchemy import create_engine

from contessa import ContessaRunner


@pytest.fixture(scope="session")
def dummy_engine():
    # should be lazy so nevermind it doesn't exists, shouldn't be used in unit tests anyway
    return create_engine("postgresql://:@postgres:5432")


@pytest.fixture(scope="session")
def dummy_contessa(dummy_engine):
    return ContessaRunner(dummy_engine)



from contessa.db import Connector
from packaging.version import parse as pv

MIGRATION_TABLE = "contessa_alembic_version"


class MigrationsResolver:
    """
    Migrations helper class for the Contessa migrations.
    """

    def __init__(self, migrations_map, package_version, url, schema):
        """
        :param migrations_map: map of package versions and their migrations.
        In form of dictionary {'0.1.4':'A', '0.1.5':'B'}
        :param package_version: the version of the package planned to be migrated
        :param url: the database url where the Alembic migration table is present or planned to be created
        :param schema: the database schema where the Alembic migration table is present or planned to be created
        """
        self.versions_migrations = migrations_map
        self.package_version = package_version
        self.url = url
        self.schema = schema
        self.conn = Connector(self.url)

    def schema_exists(self):
        """
        Check if schema with the Alembic migration table exists.
        :return: Return true if schema with the Alembic migration exists.
        """
        result = self.conn.get_records(
            f"""
            SELECT EXISTS (
               SELECT 1
               FROM   information_schema.schemata
               WHERE  schema_name = '{self.schema}'
            );
            """
        )
        return result.first()[0]

    def migrations_table_exists(self):
        """
        Check if the Alembic versions table exists.
        """
        result = self.conn.get_records(
            f"""
            SELECT EXISTS (
               SELECT 1
               FROM   information_schema.tables
               WHERE  table_schema = '{self.schema}'
               AND    table_name = '{MIGRATION_TABLE}'
           );
            """
        )
        return result.first()[0]

    def get_applied_migration(self):
        """
        Get the current applied migration in the target schema.
        """
        if self.migrations_table_exists() is False:
            return None
        version = self.conn.get_records(
            f"select * from {self.schema}.{MIGRATION_TABLE}"
        )
        return version.first()[0]

    def is_on_head(self):
        """
        Check if the current applied migration is valid for the Contessa version.
        """
        if self.migrations_table_exists() is False:
            return False
        current = self.get_applied_migration()

        fallback_package_version = self.get_fallback_version()
        return self.versions_migrations[fallback_package_version] == current

    def get_fallback_version(self):
        """
        Get fallback version in the case of non existing migration for the Contessa package version.
        Returns the last package version containing the migration.

        In the case we have this migrations versions map

        versions_migrations = {
            "0.1.4": "54f8985b0ee5",
            "0.1.5": "480e6618700d",
            "0.1.8": "3w4er8y50yyd",
            "0.1.9": "034hfa8943hr",
        }

        and we ask for the fallback version of our current version e.g. 0.1.7., we get 0.1.5. as a result because it is
        the last version with specified migration before ours.
        """
        keys = list(self.versions_migrations.keys())
        if self.package_version in self.versions_migrations.keys():
            return self.package_version
        if pv(self.package_version) < pv(keys[0]):
            return list(self.versions_migrations.keys())[0]
        if pv(self.package_version) > pv(keys[-1]):
            return list(self.versions_migrations.keys())[-1]

        result = keys[0]
        for k in keys[1:]:
            if pv(k) <= pv(self.package_version):
                result = k
            else:
                return result

    def get_migration_to_head(self):
        """
        Get the migration command for alembic. Migration command is a tupple of type of migration and migration hash.
        E.g. ('upgrade', 'dfgdfg5b0ee5') or ('downgrade', 'dfgdfg5b0ee5')
        """
        if self.is_on_head():
            return None

        fallback_version = self.get_fallback_version()

        if self.migrations_table_exists() is False:
            return "upgrade", self.versions_migrations[fallback_version]

        migrations_versions = dict(map(reversed, self.versions_migrations.items()))
        applied_migration = self.get_applied_migration()
        applied_package = migrations_versions[applied_migration]

        if pv(applied_package) < pv(fallback_version):
            return "upgrade", self.versions_migrations[fallback_version]
        if pv(applied_package) > pv(fallback_version):
            return "downgrade", self.versions_migrations[fallback_version]



from datetime import datetime, timedelta
import abc
import logging
from typing import Dict

from contessa.db import Connector
from contessa.failed_examples import ExampleSelector, default_example_selector
from contessa.models import Table


class Executor(metaclass=abc.ABCMeta):
    """
    Class that execute a rule and gives a proper kwargs to the `Rule.apply()` method.
    """

    def __init__(
        self,
        check_table: Table,
        conn: Connector,
        context: Dict,
        example_selector: ExampleSelector = default_example_selector,
    ):
        self.conn = conn
        self.check_table = check_table
        self.context = context
        self.example_selector = example_selector

    def compose_where_time_filter(self, rule):
        """
        Composes WHERE statement, which filters records by time_filter`.
        Rule attribute `time_filter` filters
        only data that were updated/created/confirmed in last 30 days.
        :return: str, WHERE `time_filter` filter statement
        """
        if rule.time_filter:
            if self.context.get("task_ts"):
                rule.time_filter.now = self.context["task_ts"]
            return rule.time_filter.sql
        return ""

    def compose_where_condition(self, rule):
        """
        Composes WHERE statement, which filters records by user-provided condition with Rule attribute `condition`.
        Very simple for now, ready for possible extensions later.
        :return: str, WHERE `condition` filter statement
        """
        condition = rule.condition
        if condition:
            return condition
        else:
            return ""

    def execute(self, rule):
        """
        Main entrypoint for Executor class. Composing kwargs specific for the executor
        (e.g. SqlExecutor) and get the results from rule.
        :param rule: Rule
        :return: AggregatedResult
        """
        kwargs = self.compose_kwargs(rule)
        results = rule.apply(**kwargs)
        return results

    @abc.abstractmethod
    def compose_kwargs(self, rule):
        raise NotImplementedError


class SqlExecutor(Executor):
    def compose_kwargs(self, rule):
        return {
            "conn": self.conn,
            "example_selector": self.example_selector,
        }


# singleton of executors
executors = None


def refresh_executors(
    check_table: Table,
    conn: Connector,
    context: Dict,
    example_selector: ExampleSelector = default_example_selector,
):
    """
    Use this to re-init the executor classes that are used to execute rules. To have right
    data from new table in Executor class.

    NOTE: This would fail if it 2 quality checks would run in 2 threads or async manner.
    Should exists a pool of executors that are mapped using schema and table_names.

    """
    global executors
    executors = {
        SqlExecutor: SqlExecutor(check_table, conn, context, example_selector),
    }
    logging.info("Successfully initialized SqlExecutor.")


def get_executor(rule):
    """
    Return instance of Executor for a specific Rule.
    :param rule: Rule
    :return: Executor
    """
    return executors[rule.executor_cls]



from abc import abstractmethod
from itertools import islice

from typing import Set, Tuple


class ExampleSelector:
    @abstractmethod
    def select_examples(self, failed_rows: Set[Tuple]) -> Set[Tuple]:
        pass


class FirstNExampleSelector(ExampleSelector):
    def __init__(self, n):
        self.n = n

    def select_examples(self, failed_rows: Set[Tuple]) -> Set[Tuple]:
        return set(islice(failed_rows, self.n))


default_example_selector = FirstNExampleSelector(10)



import logging
import re
from itertools import islice

import jinja2

from contessa.base_rules import Rule
from contessa.db import Connector
from contessa.executor import get_executor, SqlExecutor
from contessa.failed_examples import ExampleSelector, default_example_selector
from contessa.utils import AggregatedResult


class SqlRule(Rule):
    """
    Rule that executes a custom sql that is custom written.
    It can use context from Executor.
    """

    executor_cls = SqlExecutor
    only_failures_mode = False

    def get_sql_parameters(self):
        e = get_executor(self.__class__)
        return e.context

    @property
    def sql(self):
        """
        SQL query to perform on column.
        Can use context from Executor.
        """
        return f""

    def render_sql(self, sql):
        """
        Replace some parameters in query.
        :return str, formatted sql
        """
        t = jinja2.Template(sql)
        ctx = self.get_sql_parameters()
        rendered = t.render(**ctx)
        rendered = re.sub(r"%", "%%", rendered)
        return rendered

    @property
    def sql_with_where(self):
        """
        Adds `where` statement with time filter and/or user-defined condition to SQL statement.
        Could be tricky, you need to format your SQL query so WHERE statement fits to the end of it
        :return:
        """
        e = get_executor(SqlRule)
        where_clause = "WHERE "
        where_time_filter = e.compose_where_time_filter(self)
        where_condition = e.compose_where_condition(self)
        if where_time_filter == "" and where_condition == "":
            where_clause = ""
        elif where_time_filter != "" and where_condition != "":
            where_clause = f"{where_clause} {where_time_filter} AND {where_condition}"
        else:
            where_clause = f"{where_clause} {where_time_filter} {where_condition}"
        final_sql = f"{self.sql} {where_clause}"
        return self.render_sql(final_sql)

    def apply(
        self,
        conn: Connector,
        example_selector: ExampleSelector = default_example_selector,
    ):
        """
        Execute a formatted sql. Check if it returns column full of booleans representing validity that is needed
        to do a quality check. If yes, stream results and return aggregated results
        :return: AggregatedResult
        """
        sql = self.sql_with_where
        logging.debug(sql)

        failed = passed = total = 0
        failed_rows = set()

        with conn.engine.connect() as con:
            result = con.execution_options(stream_results=True).execute(sql)
            for row in result:
                if self.only_failures_mode:
                    failed += 1
                    failed_rows.add(tuple(row))
                else:
                    if not isinstance(row[0], bool) and not row[0] is None:
                        raise ValueError(
                            f"Your query for rule `{self.name}` of type `{self.type}` does not return list of booleans in column `valid`."
                        )
                    total += 1
                    if row[0] is True:
                        passed += 1
                    if row[0] is False:
                        failed += 1
                        failed_rows.add(tuple(islice(row.values(), 1, None)))

        failed_examples = example_selector.select_examples(failed_rows)

        return AggregatedResult(
            total_records=0 if self.only_failures_mode else total,
            failed=failed,
            passed=passed,
            failed_example=list(failed_examples),
        )


class OneColumnRuleSQL(SqlRule):
    def __init__(
        self, name, type, column, description, only_failures_mode=False, **kwargs
    ):
        if description == "" or description is None:
            raise TypeError("Description is mandatory")
        super().__init__(name, type, description=description, **kwargs)
        self.column = column
        self.only_failures_mode = only_failures_mode

    @property
    def attribute(self):
        return self.column

    def get_sql_parameters(self):
        context = super().get_sql_parameters()
        context.update({"target_column": self.column})
        if hasattr(self, "value"):
            context.update({"value": self.value})
        return context

    def __str__(self):
        tf = f"- {self.time_filter}" or ""
        return f"Rule {self.name} - {self.type} - {self.attribute} {tf}"


class CustomSqlRule(OneColumnRuleSQL):
    def __init__(self, name, type, column, sql, description, **kwargs):
        super().__init__(name, type, column, description, **kwargs)
        self.custom_sql = sql

    @property
    def sql(self):
        return self.custom_sql


class NotNullRule(OneColumnRuleSQL):
    def __init__(
        self, name, type, column, description="True when data is null.", **kwargs
    ):
        super().__init__(name, type, column, description=description, **kwargs)

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} IS NOT NULL,
                {{target_column}}
            FROM {{table_fullname}}
        """


class GtRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is greater than input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} > {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


class GteRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is greater or even than input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} >= {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


class NotRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is not input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} is distinct from {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


class LtRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is less than input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} < {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


class LteRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is less or even than input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} <= {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


class EqRule(OneColumnRuleSQL):
    def __init__(
        self,
        name,
        type,
        column,
        value,
        description="True when data is the same as input value.",
        **kwargs,
    ):
        super().__init__(name, type, column, description=description, **kwargs)
        self.value = value

    @property
    def sql(self):
        return """
            SELECT
                {{target_column}} IS NOT DISTINCT FROM {{value}},
                {{target_column}}
            FROM {{table_fullname}}
        """


NOT_NULL = "not_null"
NOT_COLUMN = "not_column"
GT = "gt"
GTE = "gte"
NOT = "not"
SQL = "sql"
LT = "lt"
LTE = "lte"
EQ = "eq"

RULES = {
    NOT_NULL: NotNullRule,
    GT: GtRule,
    GTE: GteRule,
    NOT: NotRule,
    SQL: CustomSqlRule,
    LT: LtRule,
    LTE: LteRule,
    EQ: EqRule,
}


def get_rule_cls(key):
    aval_rules = RULES.keys()
    if key not in aval_rules:
        raise ValueError(
            f"I dont know this kind of rule - '{key}'. "
            f"Possible rules are - {list(aval_rules)}"
        )
    return RULES[key]



"""init

Revision ID: 54f8985b0ee5
Revises: 
Create Date: 2019-11-04 14:11:49.734358

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "54f8985b0ee5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from setuptools import setup, find_packages
import contessa

setup(
    name="contessa",
    version=contessa.__version__,
    description="Data-quality framework",
    long_description=open("README-MINIMAL.rst", "r").read(),
    long_description_content_type="text/x-rst",
    license="MIT",
    author="Mario Hunka",
    author_email="mario.hunka@kiwi.com",
    url="https://github.com/kiwicom/contessa",
    packages=find_packages(),
    package_data={"contessa": ["alembic.ini", "alembic/*", "alembic/**/*"]},
    install_requires=[
        "sqlalchemy>=1.2",
        "psycopg2>=2.7",
        "jinja2>=2.10",
        "alembic>=0.8.10",
        "click>=7.0",
        "packaging>=19.2",
    ],
    tests_require=["pytest"],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": ["contessa-migrate=contessa.migration_runner:main"]
    },
    classifiers=[
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Database",
        "Topic :: Software Development :: Quality Assurance",
        "Programming Language :: Python :: 3",
    ],
)



import logging
from typing import Dict, Optional, List, Union

import jinja2
from datetime import datetime

from contessa.db import Connector
from contessa.failed_examples import default_example_selector, ExampleSelector
from contessa.models import (
    create_default_check_class,
    Table,
    ResultTable,
    ConsistencyCheck,
    CheckResult,
)
from contessa.time_filter import (
    TimeFilter,
    TimeFilterConjunction,
    TimeFilterColumn,
    parse_time_filter,
)
from contessa.utils import AggregatedResult


class ConsistencyChecker:
    """
    Checks consistency of the sync between two tables.
    """

    model_cls = ConsistencyCheck

    COUNT = "count"
    DIFF = "difference"

    def __init__(self, left_conn_uri_or_engine, right_conn_uri_or_engine=None):
        self.left_conn_uri_or_engine = left_conn_uri_or_engine
        self.left_conn = Connector(left_conn_uri_or_engine)
        if right_conn_uri_or_engine is None:
            self.right_conn_uri_or_engine = self.left_conn_uri_or_engine
            self.right_conn = self.left_conn
        else:
            self.right_conn_uri_or_engine = right_conn_uri_or_engine
            self.right_conn = Connector(right_conn_uri_or_engine)

    def run(
        self,
        method: str,
        left_check_table: Dict,
        right_check_table: Dict,
        result_table: Optional[Dict] = None,
        columns: Optional[List[str]] = None,
        time_filter: Optional[Union[str, List[Dict], TimeFilter]] = None,
        left_custom_sql: str = None,
        right_custom_sql: str = None,
        context: Optional[Dict] = None,
        example_selector: ExampleSelector = default_example_selector,
    ) -> Union[CheckResult, ConsistencyCheck]:
        if left_custom_sql and right_custom_sql:
            if columns or time_filter:
                raise ValueError(
                    "When using custom sqls you cannot change 'columns' or 'time_filter' attribute"
                )

        time_filter = parse_time_filter(time_filter)

        left_check_table = Table(**left_check_table)
        right_check_table = Table(**right_check_table)
        context = self.get_context(left_check_table, right_check_table, context)

        result = self.do_consistency_check(
            method,
            columns,
            time_filter,
            left_check_table,
            right_check_table,
            left_custom_sql,
            right_custom_sql,
            context,
            example_selector,
        )

        if result_table:
            result_table = ResultTable(**result_table, model_cls=self.model_cls)
            quality_check_class = create_default_check_class(result_table)
            self.right_conn.ensure_table(quality_check_class.__table__)
            self.upsert(quality_check_class, result)
            return result

        obj = CheckResult()
        obj.init_row_consistency(**result)
        return obj

    @staticmethod
    def get_context(
        left_check_table: Table,
        right_check_table: Table,
        context: Optional[Dict] = None,
    ) -> Dict:
        """
        Construct context to pass to executors. User context overrides defaults.
        """
        ctx_defaults = {
            "left_table_fullname": left_check_table.fullname,
            "right_table_fullname": right_check_table.fullname,
            "task_ts": datetime.now(),
        }
        if context:
            ctx_defaults.update(context)
        return ctx_defaults

    def do_consistency_check(
        self,
        method: str,
        columns: Optional[List[str]],
        time_filter: Optional[TimeFilter],
        left_check_table: Table,
        right_check_table: Table,
        left_sql: str = None,
        right_sql: str = None,
        context: Dict = None,
        example_selector: ExampleSelector = default_example_selector,
    ):
        """
        Run quality check for all rules. Use `qc_cls` to construct objects that will be inserted
        afterwards.
        """
        if not left_sql or not right_sql:
            if method == self.COUNT:
                if columns:
                    column = f"count({', '.join(columns)})"
                else:
                    column = "count(*)"
            elif method == self.DIFF:
                if columns:
                    column = ", ".join(columns)
                else:
                    # List the columns explicitly in case column order of compared tables is not the same.
                    column = ", ".join(
                        sorted(
                            self.right_conn.get_column_names(right_check_table.fullname)
                        )
                    )
            else:
                raise NotImplementedError(f"Method {method} not implemented")

        if not left_sql:
            left_sql = self.construct_default_query(
                left_check_table.fullname, column, time_filter, context
            )
        left_result = self.run_query(self.left_conn, left_sql, context)
        if not right_sql:
            right_sql = self.construct_default_query(
                right_check_table.fullname, column, time_filter, context
            )
        right_result = self.run_query(self.right_conn, right_sql, context)

        results = self.compare_results(
            left_result, right_result, method, example_selector
        )

        return {
            "check": {"type": method, "description": "", "name": "consistency",},
            "results": results,
            "left_table_name": left_check_table.fullname,
            "right_table_name": right_check_table.fullname,
            "time_filter": time_filter,
            "context": context,
        }

    def compare_results(self, left_result, right_result, method, example_selector):
        if method == self.COUNT:
            left_count = left_result[0][0]
            right_count = right_result[0][0]
            passed = min(left_count, right_count)
            failed = (left_count - passed) - (right_count - passed)
            return AggregatedResult(
                total_records=max(left_count, right_count),
                failed=failed,
                passed=passed,
            )

        elif method == self.DIFF:
            left_set = set(left_result)
            right_set = set(right_result)
            common = left_set.intersection(right_set)
            passed = len(common)
            failed = (len(left_set) - len(common)) + (len(right_set) - len(common))
            failed_examples = example_selector.select_examples(
                left_set.symmetric_difference(right_set)
            )
            return AggregatedResult(
                total_records=failed + passed,
                failed=failed,
                passed=passed,
                failed_example=list(failed_examples),
            )

        else:
            raise NotImplementedError(f"Method {method} not implemented")

    def construct_default_query(
        self,
        table_name: str,
        column: str,
        time_filter: Optional[TimeFilter],
        context: Dict,
    ):
        if time_filter:
            if context.get("task_ts"):
                time_filter.now = context["task_ts"]
            time_filter = time_filter.sql
        query = f"""
            SELECT {column}
            FROM {table_name}
            {f'WHERE {time_filter}' if time_filter else ''}
        """
        return query

    def render_sql(self, sql, context):
        """
        Replace some parameters in query.
        :return str, formatted sql
        """
        t = jinja2.Template(sql)
        rendered = t.render(**context)
        return rendered

    def run_query(self, conn: Connector, query: str, context):
        query = self.render_sql(query, context)
        logging.debug(query)
        result = [tuple(r.values()) for r in conn.get_records(query)]
        return result

    def upsert(self, dc_cls, result):
        obj = dc_cls()
        obj.init_row(**result)
        self.right_conn.upsert(
            [obj,]
        )

    def construct_automatic_time_filter(
        self, left_check_table: Dict, created_at_column=None, updated_at_column=None,
    ) -> TimeFilter:
        left_check_table = Table(**left_check_table)

        if created_at_column is None and updated_at_column is None:
            raise ValueError("Automatic time filter need at least one time column")

        since_column = updated_at_column or created_at_column
        since_sql = f"SELECT min({since_column}) FROM {left_check_table.fullname}"
        logging.info(since_sql)
        since = self.left_conn.get_records(since_sql).scalar()

        return TimeFilter(
            columns=[TimeFilterColumn(since_column, since=since),],
            conjunction=TimeFilterConjunction.AND,
        )



import pytest

from contessa.models import ResultTable, QualityCheck
from contessa.rules import GtRule, NotNullRule


def test_build_rules(dummy_contessa):
    rules = [
        {
            "name": "not_null_name",
            "type": "not_null",
            "columns": ["a", "b", "c"],
            "time_filter": "created_at",
        }
    ]
    normalized_rules = dummy_contessa.normalize_rules(rules)
    rules = dummy_contessa.build_rules(normalized_rules)
    expected = [
        NotNullRule("not_null_name", "not_null", "a", time_filter="created_at"),
        NotNullRule("not_null_name", "not_null", "b", time_filter="created_at"),
        NotNullRule("not_null_name", "not_null", "c", time_filter="created_at"),
    ]

    expected_dicts = {e.column for e in expected}
    rules_dicts = {r.column for r in rules}
    assert expected_dicts == rules_dicts


@pytest.mark.parametrize(
    "rule_def, rule_cls",
    [({"type": "not_null"}, NotNullRule), ({"type": "gt"}, GtRule)],
)
def test_pick_rule(rule_def, rule_cls, dummy_contessa):
    assert dummy_contessa.pick_rule_cls(rule_def) == rule_cls


def test_not_known_rule(dummy_contessa):
    msg = "I dont know this kind of rule .*"
    with pytest.raises(ValueError, match=msg):
        dummy_contessa.pick_rule_cls({"type": "aa"})


def test_generic_typ_qc_class(dummy_contessa):
    assert (
        dummy_contessa.get_quality_check_class(
            ResultTable("tmp", "mytable", QualityCheck)
        ).__name__
        == "TmpQualityCheckMytable"
    )


def test_generic_typ_qc_class_no_prefix(dummy_contessa):
    assert (
        dummy_contessa.get_quality_check_class(
            ResultTable("tmp", "mytable", QualityCheck)
        ).__name__
        == "TmpQualityCheckMytable"
    )
